/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
<pre>{@code
                   y
                   |
                   |
                   | v[4]
                 1 +-----------------+ v[5]=(1,1,0)
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |                    y
           |       |         |       |                    |
           |       |         |       | v[1]               |
           |  v[0] +---------|-------+------> x           |
           |      /          |      /1                    |
           |    /            |    /                       +-----> x
           |  /              |  /                        /
           |/                |/                         /
         1 +-----------------+                         /
         / v[3]=(0,0,1)      v[2]=(1,0,1)            z
       /
     /
    z
}</pre>
   Render a single wireframe cube.
*/
@SuppressWarnings("serial")
public class InteractiveCube_R13 extends InteractiveFrame
{
   private boolean render_z_buffer = false;

   private boolean doBackFaceCulling = true;
   private boolean frontFacingIsCCW = true;

   private double eyeX = 0.0;
   private double eyeY = 0.0;
   private double eyeZ = 2.0;
   private double eyeRotX = 0.0;
   private double eyeRotY = 0.0;
   private double eyeRotZ = 0.0;

   private boolean letterbox = false;
   private double aspectRatio = 1.0;
   private double near   =  1.0;
   private double left   = -1.0;
   private double right  =  1.0;
   private double bottom = -1.0;
   private double top    =  1.0;
   private boolean showCamera = false;

   private boolean showMatrix = false;
   private double xTranslation = 0.0;
   private double yTranslation = 0.0;
   private double zTranslation = 0.0;
   private double xRotation = 0.0;
   private double yRotation = 0.0;
   private double zRotation = 0.0;
   private double scale = 1.0;

   private Scene scene;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public InteractiveCube_R13(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create a Model object to hold the geometry.
      Model model = new Model();

      // Create the geometry for the Model.
      Vertex v0, v1, v2, v3, v4, v5, v6, v7;
      v0 = new Vertex(0.0, 0.0, 0.0); // four vertices around the bottom face
      v1 = new Vertex(1.0, 0.0, 0.0);
      v2 = new Vertex(1.0, 0.0, 1.0);
      v3 = new Vertex(0.0, 0.0, 1.0);
      v4 = new Vertex(0.0, 1.0, 0.0); // four vertices around the top face
      v5 = new Vertex(1.0, 1.0, 0.0);
      v6 = new Vertex(1.0, 1.0, 1.0);
      v7 = new Vertex(0.0, 1.0, 1.0);

      // Give each vertex a different color.
      v0.setColorRandom();
      v1.setColorRandom();
      v2.setColorRandom();
      v3.setColorRandom();
      v4.setColorRandom();
      v5.setColorRandom();
      v6.setColorRandom();
      v7.setColorRandom();

      // Add the geometry to the model.
      model.addTriangle(v0, v1, v2);  // bottom face
      model.addTriangle(v0, v2, v3);  // bottom face
      model.addTriangle(v7, v6, v5);  // top face
      model.addTriangle(v7, v5, v4);  // top face
      model.addTriangle(v0, v4, v5);  // back face
      model.addTriangle(v0, v5, v1);  // back face
      model.addTriangle(v3, v2, v6);  // front face
      model.addTriangle(v3, v6, v7);  // front face
      model.addTriangle(v0, v3, v7);  // left face
      model.addTriangle(v0, v7, v4);  // left face
      model.addTriangle(v1, v5, v6);  // right face
      model.addTriangle(v1, v6, v2);  // right face
/*
      // Or we can add the geometry to the model in this more efficient way.
      model.addVertex(v0, v1, v2, v3, v4, v5, v6, v7);
      model.addTriangle(0, 1, 2);  // bottom face
      model.addTriangle(0, 2, 3);  // bottom face
      model.addTriangle(7, 6, 5);  // top face
      model.addTriangle(7, 5, 4);  // top face
      model.addTriangle(0, 4, 5);  // back face
      model.addTriangle(0, 5, 1);  // back face
      model.addTriangle(3, 2, 6);  // front face
      model.addTriangle(3, 6, 7);  // front face
      model.addTriangle(0, 3, 7);  // left face
      model.addTriangle(0, 7, 4);  // left face
      model.addTriangle(1, 5, 6);  // right face
      model.addTriangle(1, 6, 2);  // right face
*/
      // Create a Position for this Model.
      Position position = new Position(model);

      // Create the Scene object that we shall render.
      scene = new Scene();

      // Add the Position (and its Model) to the Scene.
      scene.addPosition(position);

      // Set up the camera's location and orientation.
      scene.camera.viewTranslate(eyeX, eyeY, eyeZ);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      int keyCode = e.getKeyCode();
      // Only handle the four arrow keys.
      if (KeyEvent.VK_UP == keyCode   || KeyEvent.VK_DOWN == keyCode
       || KeyEvent.VK_LEFT == keyCode || KeyEvent.VK_RIGHT == keyCode)
      {
         if (KeyEvent.VK_UP == keyCode && e.getModifiers() != 0 )
         {
            eyeRotX += 1;
         }
         else if (KeyEvent.VK_DOWN == keyCode && e.getModifiers() != 0 )
         {
            eyeRotX -= 1;
         }
         else if (KeyEvent.VK_LEFT == keyCode && e.getModifiers() != 0 )
         {
            eyeRotY += 1;
         }
         else if (KeyEvent.VK_RIGHT == keyCode && e.getModifiers() != 0 )
         {
            eyeRotY -= 1;
         }
         else if (KeyEvent.VK_UP == keyCode)
         {
            eyeY += 0.1;
         }
         else if (KeyEvent.VK_DOWN == keyCode)
         {
            eyeY -= 0.1;
         }
         else if (KeyEvent.VK_LEFT == keyCode)
         {
            eyeX -= 0.1;
         }
         else if (KeyEvent.VK_RIGHT == keyCode)
         {
            eyeX += 0.1;
         }

         // Set up the camera's location and orientation.
         scene.camera.view2Identity();
         scene.camera.viewTranslate(eyeX, eyeY, eyeZ);
         scene.camera.viewRotateX(eyeRotX);
         scene.camera.viewRotateY(eyeRotY);

         if (showCamera)
         {
            System.out.println("eyeRotX = " + eyeRotX
                           + ", eyeRotY = " + eyeRotY);
            System.out.print( scene.camera.viewMatrix );
         }

         // Render again.
         setupViewport();
      }
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
       //Clip.debug = ! Clip.debug;
       //Rasterize.debug = ! Rasterize.debug;
      }
      else if (' ' == c)
      {
         render_z_buffer = ! render_z_buffer;
      }
      else if ('g' == c)
      {
         doBackFaceCulling = ! doBackFaceCulling;
         // Traverse the whole scene graph.
         for (Position position : scene.positionList)
            setBackFaceCulling(position, doBackFaceCulling);
         System.out.print("Back face culling is ");
         System.out.println(doBackFaceCulling ? "on" : "off");
      }
      else if ('G' == c)
      {
         frontFacingIsCCW = ! frontFacingIsCCW;
         // Traverse the whole scene graph.
         for (Position position : scene.positionList)
            setFrontFacingIsCCW(position, frontFacingIsCCW);
         System.out.print("Front face is ");
         System.out.println(frontFacingIsCCW ? "CCW" : "CW");
      }
      else if ('n' == c || 'N' == c)
      {
         // Move the camera's near plane.
         if ('n' == c)
         {
            near -= 0.01;
         }
         else
         {
            near += 0.01;
         }
      }
      else if ('r' == c || 'R' == c)
      {
         // Change the aspect ratio of the camera's view rectangle.
         if ('r' == c)
         {
            aspectRatio -= 0.1;
         }
         else
         {
            aspectRatio += 0.1;
         }

         // Adjust right and left.
         // (Keep the vertical field-of-view fixed.)
         right =  top * aspectRatio;
         left  = -right;
      }
      else if ('o' == c || 'O' == c)
      {
         // Change left, right, bottom, and top.
         // (Keep the aspect ratio fixed.)
         if ('o' == c)
         {
            left   += 0.1 * aspectRatio;
            right  -= 0.1 * aspectRatio;
            bottom += 0.1;
            top    -= 0.1;
         }
         else
         {
            left   -= 0.1 * aspectRatio;
            right  += 0.1 * aspectRatio;
            bottom -= 0.1;
            top    += 0.1;
         }
      }
      else if ('l' == c)
      {
         letterbox = ! letterbox;
      }
      else if ('c' == c)
      {
         // Change the solid random color of the cube.
         scene.positionList.get(0).model.setColorRandom();
      }
      else if ('e' == c)
      {
         // Change the random color of each vertex.
         for (Vertex v : scene.positionList.get(0).model.vertexList)
         {
            v.setColorRandom();
         }
      }
      else if ('E' == c)
      {
         // Change the random color of each line segment.
         for (Primitive p : scene.positionList.get(0).model.primitiveList)
         {
            p.setColorRandom();
         }
      }
      else if ('a' == c)
      {
         Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
      }
      else if ('M' == c)
      {
         showCamera = ! showCamera;
      }
      else if ('m' == c)
      {
         showMatrix = ! showMatrix;
      }
      else if ('s' == c) // Scale the cube 10% smaller.
      {
         scale /= 1.1;
      }
      else if ('S' == c) // Scale the cube 10% larger.
      {
         scale *= 1.1;
      }
      else if ('x' == c)
      {
         xTranslation -= 0.1;
      }
      else if ('X' == c)
      {
         xTranslation += 0.1;
      }
      else if ('y' == c)
      {
         yTranslation -= 0.1;
      }
      else if ('Y' == c)
      {
         yTranslation += 0.1;
      }
      else if ('z' == c)
      {
         zTranslation -= 0.1;
      }
      else if ('Z' == c)
      {
         zTranslation += 0.1;
      }
      else if ('u' == c)
      {
         xRotation -= 2.0;
      }
      else if ('U' == c)
      {
         xRotation += 2.0;
      }
      else if ('v' == c)
      {
         yRotation -= 2.0;
      }
      else if ('V' == c)
      {
         yRotation += 2.0;
      }
      else if ('w' == c)
      {
         zRotation -= 2.0;
      }
      else if ('W' == c)
      {
         zRotation += 2.0;
      }

      // Set the model-to-world transformation matrix.
      // The order of the transformations is important!
      Position cube_p = scene.positionList.get(0);
      cube_p.matrix2Identity();
      // Move the model relative to its new position.
      cube_p.translate(xTranslation, yTranslation, zTranslation);
      cube_p.rotateX(xRotation);
      cube_p.rotateY(yRotation);
      cube_p.rotateZ(zRotation);
      cube_p.scale(scale);

      // Set up the camera's view volume.
      if (scene.camera.perspective)
      {
         scene.camera.projPerspective(left, right, bottom, top, near);
      }
      else
      {
         scene.camera.projOrtho(left, right, bottom, top);
      }

      if (showCamera && ('M'==c
           ||'n'==c||'N'==c||'o'==c||'O'==c||'r'==c||'R'==c||'p'==c))
      {
         System.out.print( scene.camera );
      }

      if (showMatrix && ('m'==c
           ||'s'==c||'x'==c||'y'==c||'z'==c||'u'==c||'v'==c||'w'==c
           ||'S'==c||'X'==c||'Y'==c||'Z'==c||'U'==c||'V'==c||'W'==c))
      {
         System.out.println("xRot = " + xRotation
                        + ", yRot = " + yRotation
                        + ", zRot = " + zRotation);
         System.out.print( cube_p.modelMatrix );
      }

      // Render again.
      setupViewport();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);

      // Render again.
      setupViewport();
   }


   // Get in one place the code to set up the viewport.
   private void setupViewport()
   {
      // Render again.
      // Get the size of the FrameBuffer.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      int w = fb.getWidthFB();
      int h = fb.getHeightFB();
      // Create a viewport with the correct aspect ratio.
      if ( letterbox )
      {
         if ( aspectRatio <= w/(double)h )
         {
            int width = (int)(h*aspectRatio);
            int xOffset = (w - width)/2;
            fb.setViewport(xOffset, 0, width, h);
         }
         else
         {
            int height = (int)(w/aspectRatio);
            int yOffset = (h - height)/2;
            fb.setViewport(0, yOffset, w, height);
         }
         fb.clearFB(Color.darkGray);
         fb.clearVP(Color.black);
      }
      else // the viewport is the whole framebuffer
      {
         fb.clearFB(Color.black);
         fb.setViewport();
      }
      Pipeline.render(scene, fb);
      if ( render_z_buffer )
      {
         FrameBuffer zBuf = this.fbp.getFrameBuffer().convertZB2FB();
         this.fbp.getFrameBuffer().setViewport(0, 0, zBuf);
      }
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // InteractiveCube_R13 constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new InteractiveCube_R13("Renderer 13", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the cube along the x, y, z axes.");
      System.out.println("Use the u/U, v/V, w/W, keys to rotate the cube around the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the cube.");
      System.out.println("Use the 'c' key to change the solid cube color.");
      System.out.println("Use the 'e' key to change the random vertex colors.");
      System.out.println("Use the 'E' key to change the random line segment colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the n/N keys to move the camera's near plane.");
      System.out.println("Use the o/O keys to change the size of the camera's view rectangle.");
      System.out.println("Use the r/R keys to change the aspect ratio of the camera's view rectangle.");
      System.out.println("Use the 'l' key to toggle letterboxing on and off.");
      System.out.println("Use the arrow keys to move the camera location left/right and up/down.");
      System.out.println("Use CTRL arrow keys to rotate the camera left/right and up/down.");
      System.out.println("Use the 'm' key to toggle showing the Model transformation matrix.");
      System.out.println("Use the 'M' key to toggle showing the Camera's three matrices.");
      System.out.println("Use the 'g' key to toggle back face culling.");
      System.out.println("Use the 'G' key to toggle CW vs CCW.");
      System.out.println("Use the ' ' key to toggle between the image buffer and z-buffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   private static void setBackFaceCulling(Position p, boolean doBackFaceCulling)
   {
      if (null != p.model)
      {
         p.model.doBackFaceCulling = doBackFaceCulling;
         for (Position p2 : p.model.nestedPositions)
         {
            setBackFaceCulling(p2, doBackFaceCulling);
         }
      }
      for (Position p3 : p.nestedPositions)
      {
         setBackFaceCulling(p3, doBackFaceCulling);
      }
   }


   private static void setFrontFacingIsCCW(Position p, boolean frontFacingIsCCW)
   {
      if (null != p.model)
      {
         p.model.frontFacingIsCCW = frontFacingIsCCW;
         for (Position p2 : p.model.nestedPositions)
         {
            setFrontFacingIsCCW(p2, frontFacingIsCCW);
         }
      }
      for (Position p3 : p.nestedPositions)
      {
         setFrontFacingIsCCW(p3, frontFacingIsCCW);
      }
   }
}
