/*

*/

package renderer.models_F;
import  renderer.scene.*;

/**
   Create a wireframe model of a cube with its center
   at the origin, having edge length 2, and with its
   corners at {@code (1, 1, 1)}.
<p>
   This version of the cube model has each face of
   the cube cut up by a triangle fan.
<p>
   Here is a picture showing how the cube's eight corners
   are labeled.
<pre>{@code
                  v[4]
                   +-----------------+ v[5]
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |               y
           |       |         |       |               |
           |       |         |       |               |
           |  v[0] +---------|-------+ v[1]          |
           |      /          |      /                |
           |    /            |    /                  +----. x
           |  /              |  /                   /
           |/                |/                    /
           +-----------------+                    /
          v[3]              v[2]                 z
}</pre>

   @see Cube
   @see Cube2
   @see Cube3
*/
public class Cube4 extends Model
{
   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}. and
      with a triangle fan of four triangles in each face.
   */
   public Cube4( )
   {
      this(1, 1, 1);
   }


   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}, and
      with each of the cube's faces containing a triangle fan
      with the given number of triangles along each of the x,
      y, and z directions.
      <p>
      There must be at least one triangle along each direction.

      @param xCount  number of triangles along the x-direction
      @param yCount  number of triangles along the y-direction
      @param zCount  number of triangles along the z-direction
   */
   public Cube4(int xCount, int yCount, int zCount)
   {
      if (xCount < 1) xCount = 1;
      if (yCount < 1) yCount = 1;
      if (zCount < 1) zCount = 1;

      double xStep = 2.0 / xCount;
      double yStep = 2.0 / yCount;
      double zStep = 2.0 / zCount;

      Vertex cFront  = new Vertex(0,  0,  1); // center front
      Vertex cBack   = new Vertex(0,  0, -1); // center back
      Vertex cTop    = new Vertex(0,  1,  0); // top center
      Vertex cBottom = new Vertex(0, -1,  0); // center bottom
      Vertex rCenter = new Vertex( 1, 0,  0); // center right
      Vertex lCenter = new Vertex(-1, 0,  0); // center left

      // Triangles along all four edges parallel to the x-axis.
      double x = -1.0;
      for (int i = 0; i < xCount; i++)
      {  // front face, top and bottom edges
         addFace(new Vertex(x,  1,  1), cFront, new Vertex(x+xStep,  1,  1));
         addFace(new Vertex(x, -1,  1), new Vertex(x+xStep, -1,  1), cFront);
         // back face, top and bottom edges
         addFace(new Vertex(x,  1, -1), new Vertex(x+xStep,  1, -1), cBack);
         addFace(new Vertex(x, -1, -1), cBack, new Vertex(x+xStep, -1, -1));
         // top face, front and back edges
         addFace(new Vertex(x,  1,  1), new Vertex(x+xStep,  1,  1), cTop);
         addFace(new Vertex(x,  1, -1), cTop, new Vertex(x+xStep,  1, -1));
         // bottom face, front and back edges
         addFace(new Vertex(x, -1,  1), cBottom, new Vertex(x+xStep, -1,  1));
         addFace(new Vertex(x, -1, -1), new Vertex(x+xStep, -1, -1), cBottom);
         x += xStep;
      }

      // Triangles along all four edges parallel to the y-axis.
      double y = -1.0;
      for (int i = 0; i < yCount; i++)
      {  // front face, right and left edges
         addFace(new Vertex( 1, y,  1), new Vertex( 1, y+yStep,  1), cFront);
         addFace(new Vertex(-1, y,  1), cFront,  new Vertex(-1, y+yStep,  1));
         // back face, right and left edges
         addFace(new Vertex( 1, y, -1), cBack,   new Vertex( 1, y+yStep, -1));
         addFace(new Vertex(-1, y, -1),   new Vertex(-1, y+yStep, -1), cBack);
         // right face, front and back edges
         addFace(new Vertex( 1, y,  1), rCenter, new Vertex( 1, y+yStep,  1));
         addFace(new Vertex( 1, y, -1), new Vertex( 1, y+yStep, -1), rCenter);
         // left face, front and back edges
         addFace(new Vertex(-1, y,  1), new Vertex(-1, y+yStep,  1), lCenter);
         addFace(new Vertex(-1, y, -1), lCenter, new Vertex(-1, y+yStep, -1));
         y += yStep;
      }

      // Triangles along all four edges parallel to the z-axis.
      double z = -1.0;
      for (int i = 0; i < zCount; i++)
      {  // top face, right and left edges
         addFace(new Vertex( 1,  1, z), cTop, new Vertex( 1,  1, z+zStep));
         addFace(new Vertex(-1,  1, z), new Vertex(-1,  1, z+zStep), cTop);
         // bottom face, right and left edges
         addFace(new Vertex( 1, -1, z), new Vertex( 1, -1, z+zStep), cBottom);
         addFace(new Vertex(-1, -1, z), cBottom, new Vertex(-1, -1, z+zStep));
         // right face, top and bottom edges
         addFace(new Vertex( 1,  1, z), new Vertex( 1,  1, z+zStep), rCenter);
         addFace(new Vertex( 1, -1, z), rCenter, new Vertex( 1, -1, z+zStep));
         // left face, top and bottom edges
         addFace(new Vertex(-1,  1, z), lCenter, new Vertex(-1,  1, z+zStep));
         addFace(new Vertex(-1, -1, z), new Vertex(-1, -1, z+zStep), lCenter);
         z += zStep;
      }
   }
}//Cube4
