/*

*/

package renderer.models_F;
import  renderer.scene.*;

/**
   Create a wireframe model of a disk
   in the xy-plane centered at the origin.
<p>
   See <a href="https://en.wikipedia.org/wiki/Disk_(mathematics)" target="_top">
                https://en.wikipedia.org/wiki/Disk_(mathematics)</a>

   @see DiskSector
*/
public class Disk extends Model
{
   /**
      Create a disk in the xy-plane with radius 1,
      with 12 spokes coming out of the center.
   */
   public Disk( )
   {
      this(1.0, 1, 12);
   }


   /**
      Create a disk in the xy-plane with radius
      {@code r}, with 12 spokes coming out of the
      center.

      @param r  radius of the disk
   */
   public Disk(double r)
   {
      this(r, 1, 12);
   }


   /**
      Create a disk in the xy-plane with radius
      {@code r}, with {@code k} spokes coming out
      of the center, and with {@code n} concentric
      circles around the disk.
   <p>
      Notice that if there are {@code k} spokes, then each
      circle around the center will have {@code k} line segments.
      If there are {@code n} concentric circles around the
      center, then each spoke will have {@code n} line segments.
   <p>
      There must be at least three spokes and at least
      one concentric circle.

      @param r  radius of the disk
      @param n  number of concentric circles
      @param k  number of spokes in the disk
   */
   public Disk(double r, int n, int k)
   {
      super();

      if (n < 1) n = 1;
      if (k < 3) k = 3;

      // Create the disk's geometry.

      double deltaR = r / n;
      double deltaTheta = 2 * Math.PI / k;

      // An array of vertices to be used to create line segments.
      Vertex[][] v = new Vertex[n][k];

      // Create all the vertices.
      for (int j = 0; j < k; j++) // choose a spoke (an angle)
      {
         double c = Math.cos(j * deltaTheta);
         double s = Math.sin(j * deltaTheta);
         for (int i = 0; i < n; i++) // move along the spoke
         {
            double ri = (i + 1) * deltaR;
            v[i][j] = new Vertex( ri * c,
                                  ri * s,
                                  0 );
         }
      }
      Vertex center = new Vertex(0,0,0);

      // Create the triangle fan at the center.
      for (int j = 0; j < k-1; j++)
      {
         addFace(center, v[0][j], v[0][j+1]);
      }
      addFace(center, v[0][k-1], v[0][0]);

      // Create all the square strips.
      for (int i = 0; i < n-1; i++)
      {
         for (int j = 0; j < k-1; j++)
         {
            addFace(v[i][j], v[i+1][j], v[i+1][j+1], v[i][j+1]);
         }
         addFace(v[i][k-1], v[i+1][k-1], v[i+1][0], v[i][0]);
      }
   }
}//Disk
