/*

*/

package renderer.models_F;
import  renderer.scene.*;

/**
   Create a wireframe model of a regular tetrahedron
   with its center at the origin, having edge length
   {@code 2*sqrt(2)}, and with its vertices at corners
   of the cube with vertices {@code (1, 1, 1)}.
<p>
   See <a href="http://en.wikipedia.org/wiki/Tetrahedron" target="_top">
                http://en.wikipedia.org/wiki/Tetrahedron</a>

   @see Cube
   @see Octahedron
   @see Icosahedron
   @see Dodecahedron
*/
public class Tetrahedron extends Model
{
   /**
      Create a regular tetrahedron with its center at
      the origin, having edge length {@code 2*sqrt(2)},
      and with its vertices at corners of the cube with
      vertices {@code (1, 1, 1)}.
   */
   public Tetrahedron()
   {
      this(false);
   }


   /**
      Create a regular tetrahedron or its dual tetrahedron
      (the dual of a tetrahedron is another tetrahedron).
   <p>
      <a href="https://en.wikipedia.org/wiki/Tetrahedron#Regular_tetrahedron" target="_top">
               https://en.wikipedia.org/wiki/Tetrahedron#Regular_tetrahedron</a>
   <p>
      The combination of these two dual tetrahedrons is a stellated octahedron.
   <p>
      <a href="https://en.wikipedia.org/wiki/Stellated_octahedron" target="_top">
               https://en.wikipedia.org/wiki/Stellated_octahedron</a>

      @param dual  choose between the two dual tetrahedrons
   */
   public Tetrahedron(boolean dual)
   {
      super();

      // Create the tetrahedron's geometry.
      // It has 4 vertices, 6 edges, and 4 faces.
      Vertex[] v = new Vertex[4];
      if ( ! dual)
      {
         v[0] = new Vertex( 1,  1,  1);
         v[1] = new Vertex(-1,  1, -1);
         v[2] = new Vertex( 1, -1, -1);
         v[3] = new Vertex(-1, -1,  1);
      }
      else // create the dual tetrahedron by
      {    // inverting the coordinates given above
         v[0] = new Vertex(-1, -1, -1);
         v[1] = new Vertex( 1, -1,  1);
         v[2] = new Vertex(-1,  1,  1);
         v[3] = new Vertex( 1,  1, -1);
      }

      // Create four triangular faces.
      // Make sure the faces are all
      // oriented in the same way!
      // Remember that front facing
      // should be counterclockwise
      // (and back facing is clockwise).
      addFace(v[3], v[1], v[0]);
      addFace(v[3], v[2], v[1]);
      addFace(v[3], v[0], v[2]);
      addFace(v[0], v[1], v[2]);
   }
}//Tetrahedron
