/*

*/

package renderer.models_T;
import  renderer.scene.*;

/**
   Create an x and y axis in the xy-plane, along with "tick marks".
*/
public class Axes2D extends Model
{
   /**
      Create an x and y axis from -1 to +1 on each axis.
   */
   public Axes2D( )
   {
      this(-1, 1, -1, 1, 0.0, 5, 5);
   }


   /**
      Create an x-axis from {@code xMin} to {@code xMax}
      and a y-axis from {@code yMin} to {@code yMax}.

      @param xMin    left end point for the x-axis
      @param xMax    right end point for the x-axis
      @param yMin    bottom end point for the y-axis
      @param yMax    top end point for the y-axis
      @param xMarks  number of evenly spaced tick marks on the x-axis
      @param yMarks  number of evenly spaced tick marks on the y-axis
   */
   public Axes2D(double xMin, double xMax, double yMin, double yMax,
                 int xMarks, int yMarks)
   {
      this(xMin, xMax, yMin, yMax, 0.0, xMarks, yMarks);
   }


   /**
      Create an x-axis from {@code xMin} to {@code xMax}
      and a y-axis from {@code yMin} to {@code yMax}.
   <p>
      The {@code z} parameter is so that we can put the axis just above
      or just below the xy-plane (say {@code z=0.01} or {@code z=-0.01}).
      This way, the axes can be just in front of or just behind whatever
      is being drawn in the xy-plane.

      @param xMin    left end point for the x-axis
      @param xMax    right end point for the x-axis
      @param yMin    bottom end point for the y-axis
      @param yMax    top end point for the y-axis
      @param z       offset of the axes away from the xy-plane
      @param xMarks  number of evenly spaced tick marks on the x-axis
      @param yMarks  number of evenly spaced tick marks on the y-axis
   */
   public Axes2D(double xMin, double xMax, double yMin, double yMax,
                 double z, int xMarks, int yMarks)
   {
      super();

       // x-axis
      addLineSegment(new Vertex(xMin, 0, z),
                     new Vertex(xMax, 0, z));

       // y-axis
      addLineSegment(new Vertex(0, yMin, z),
                     new Vertex(0, yMax, z));

       // Put evenly spaced tick marks on the x-axis.
      double xDelta = (xMax - xMin)/xMarks;
      double yDelta = (yMax - yMin)/50;
      for (double x = xMin; x <= xMax; x += xDelta)
      {
         addLineSegment(new Vertex(x,  yDelta/2, z),
                        new Vertex(x, -yDelta/2, z));
      }

       // Put evenly spaced tick marks on the y-axis.
      yDelta = (yMax - yMin)/yMarks;
      xDelta = (xMax - xMin)/50;
      for (double y = yMin; y <= yMax; y += yDelta)
      {
         addLineSegment(new Vertex( xDelta/2, y, z),
                        new Vertex(-xDelta/2, y, z));
      }
   }
}//Axes2D
