/*

*/

package renderer.models_T;
import  renderer.scene.*;

/**
   Create a wireframe model of a cube with its center
   at the origin, having edge length 2, and with its
   corners at {@code (1, 1, 1)}.
<p>
   This version of the cube model has each face of
   the cube cut up by an n by m grid of lines.
<p>
   Here is a picture showing how the cube's eight corners
   are labeled.
<pre>{@code
                  v[4]
                   +-----------------+ v[5]
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |               y
           |       |         |       |               |
           |       |         |       |               |
           |  v[0] +---------|-------+ v[1]          |
           |      /          |      /                |
           |    /            |    /                  +----. x
           |  /              |  /                   /
           |/                |/                    /
           +-----------------+                    /
          v[3]              v[2]                 z
}</pre>

   @see Cube
   @see Cube3
   @see Cube4
*/
public class Cube2 extends Model
{
   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}. and
      with two perpendicular grid lines going across the middle
      of each of the cube's faces.
   */
   public Cube2( )
   {
      this(1, 1, 1);
   }


   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}, and
      with each of the cube's faces containing the given number
      of grid lines parallel to the x, y, and z directions.

      @param xGrid  number of grid lines perpendicular to the x-axis
      @param yGrid  number of grid lines perpendicular to the y-axis
      @param zGrid  number of grid lines perpendicular to the z-axis
   */
   public Cube2(int xGrid, int yGrid, int zGrid)
   {
      if (xGrid < 0) xGrid = 0;
      if (yGrid < 0) yGrid = 0;
      if (zGrid < 0) zGrid = 0;

      double xStep = 2.0 / (1 + xGrid);
      double yStep = 2.0 / (1 + yGrid);
      double zStep = 2.0 / (1 + zGrid);

      // Create the top and bottom sides of the cube.
      // An array of vertices to be used to create the triangles.
      Vertex[][] vT = new Vertex[xGrid+2][zGrid+2];
      Vertex[][] vB = new Vertex[xGrid+2][zGrid+2];
      // Create all the vertices.
      for (int i = 0; i <  xGrid + 2; i++)
      {
         for (int j = 0; j < zGrid + 2; j++)
         {
            vT[i][j] = new Vertex(-1 + i * xStep,  1, -1 + j * zStep);
            vB[i][j] = new Vertex(-1 + i * xStep, -1, -1 + j * zStep);
         }
      }
      // Create the triangles.
      for (int i = 0; i < xGrid + 1; i++)
      {
         for (int j = 0; j < zGrid + 1; j++)
         {
            addTriangle( vT[i  ][j  ], vT[i  ][j+1], vT[i+1][j+1] );
            addTriangle( vT[i+1][j+1], vT[i+1][j  ], vT[i  ][j  ] );
            addTriangle( vB[i  ][j  ], vB[i+1][j+1], vB[i  ][j+1] );
            addTriangle( vB[i+1][j+1], vB[i  ][j  ], vB[i+1][j  ] );
         }
      }

      // Create the front and back sides of the cube.
      // An array of vertices to be used to create the triangles.
      Vertex[][] vF = new Vertex[yGrid+2][xGrid+2];
                 vB = new Vertex[yGrid+2][xGrid+2];
      // Create all the vertices.
      for (int i = 0; i <  yGrid + 2; i++)
      {
         for (int j = 0; j < xGrid + 2; j++)
         {
            vF[i][j] = new Vertex(-1 + j * xStep, -1 + i * yStep,  1);
            vB[i][j] = new Vertex(-1 + j * xStep, -1 + i * yStep, -1);
         }
      }
      // Create the triangles.
      for (int i = 0; i < yGrid + 1; i++)
      {
         for (int j = 0; j < xGrid + 1; j++)
         {
            addTriangle( vF[i  ][j  ], vF[i  ][j+1], vF[i+1][j+1] );
            addTriangle( vF[i+1][j+1], vF[i+1][j  ], vF[i  ][j  ] );
            addTriangle( vB[i  ][j  ], vB[i+1][j+1], vB[i  ][j+1] );
            addTriangle( vB[i+1][j+1], vB[i  ][j  ], vB[i+1][j  ] );
         }
      }

      // Create the right and left sides of the cube.
      // An array of vertices to be used to create the triangles.
      Vertex[][] vR = new Vertex[yGrid+2][zGrid+2];
      Vertex[][] vL = new Vertex[yGrid+2][zGrid+2];
      // Create all the vertices.
      for (int i = 0; i < yGrid + 2; i++)
      {
         for (int j = 0; j < zGrid + 2; j++)
         {
            vR[i][j] = new Vertex( 1, -1 + i * yStep, -1 + j * zStep);
            vL[i][j] = new Vertex(-1, -1 + i * yStep, -1 + j * zStep);
         }
      }
      // Create the triangles.
      for (int i = 0; i < yGrid + 1; i++)
      {
         for (int j = 0; j < zGrid + 1; j++)
         {
            addTriangle( vR[i  ][j  ], vR[i+1][j+1], vR[i  ][j+1] );
            addTriangle( vR[i+1][j+1], vR[i  ][j  ], vR[i+1][j  ] );
            addTriangle( vL[i  ][j  ], vL[i  ][j+1], vL[i+1][j+1] );
            addTriangle( vL[i+1][j+1], vL[i+1][j  ], vL[i  ][j  ] );
         }
      }
   }
}//Cube2
