/*

*/

package renderer.models_T;
import  renderer.scene.*;

/**
   Create a wireframe model of a cube with its center
   at the origin, having edge length 2, and with its
   corners at {@code (1, 1, 1)}.
<p>
   This version of the cube model has the top and bottom
   faces of the cube cut up by a triangle fan and the
   front, back, right, and left faces cut up by a grid
   of perpendicular lines..
<p>
   Here is a picture showing how the cube's eight corners
   are labeled.
<pre>{@code
                  v[4]
                   +-----------------+ v[5]
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |               y
           |       |         |       |               |
           |       |         |       |               |
           |  v[0] +---------|-------+ v[1]          |
           |      /          |      /                |
           |    /            |    /                  +----. x
           |  /              |  /                   /
           |/                |/                    /
           +-----------------+                    /
          v[3]              v[2]                 z
}</pre>

   @see Cube
   @see Cube2
   @see Cube4
*/
public class Cube3 extends Model
{
   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}. with
      a triangle fan of eight triangles in top and bottom faces,
      and two perpendicular lines cutting each of the front,
      back, right, and left faces.
   */
   public Cube3( )
   {
      this(2, 1, 2);
   }


   /**
      Create a cube with its center at the origin, having edge
      length 2, with its corners at {@code (1, 1, 1)}, and
      with the top and bottom faces containing a triangle fan
      with the given number of triangles along each of the x,
      and z directions.
      <p>
      There must be at least one triangle along the x and z directions.

      @param xCount  number of triangles along the x-direction
      @param yGrid   number of grid lines perpendicular to the y-axis
      @param zCount  number of triangles along the z-direction
   */
   public Cube3(int xCount, int yGrid, int zCount)
   {
      if (xCount < 1) xCount = 1;
      if (yGrid  < 0) yGrid  = 0;
      if (zCount < 1) zCount = 1;

      double xStep = 2.0 / xCount;
      double yStep = 2.0 / (1 + yGrid);
      double zStep = 2.0 / zCount;

      Vertex cTop    = new Vertex(0,  1,  0); // top center
      Vertex cBottom = new Vertex(0, -1,  0); // center bottom

      // Create the top and bottom sides of the cube.
      // Triangles along all four edges parallel to the x-axis.
      double x = -1.0;
      for (int i = 0; i < xCount; i++)
      {  // top face, front and back edges
         addTriangle(new Vertex(x,  1,  1), new Vertex(x+xStep,  1,  1), cTop);
         addTriangle(new Vertex(x,  1, -1), cTop, new Vertex(x+xStep,  1, -1));
         // bottom face, front and back edges
         addTriangle(new Vertex(x, -1,  1), cBottom, new Vertex(x+xStep, -1,  1));
         addTriangle(new Vertex(x, -1, -1), new Vertex(x+xStep, -1, -1), cBottom);
         x += xStep;
      }
      // Triangles along all four edges parallel to the z-axis.
      double z = -1.0;
      for (int i = 0; i < zCount; i++)
      {  // top face, right and left edges
         addTriangle(new Vertex( 1,  1, z), cTop, new Vertex( 1,  1, z+zStep));
         addTriangle(new Vertex(-1,  1, z), new Vertex(-1,  1, z+zStep), cTop);
         // bottom face, right and left edges
         addTriangle(new Vertex( 1, -1, z), new Vertex( 1, -1, z+zStep), cBottom);
         addTriangle(new Vertex(-1, -1, z), cBottom, new Vertex(-1, -1, z+zStep));
         z += zStep;
      }

      // Create the front and back sides of the cube.
      // An array of vertices to be used to create the triangles.
      Vertex[][] vF = new Vertex[yGrid+2][xCount+1];
      Vertex[][] vB = new Vertex[yGrid+2][xCount+1];
      // Create all the vertices.
      for (int i = 0; i <  yGrid + 2; i++)
      {
         for (int j = 0; j < xCount + 1; j++)
         {
            vF[i][j] = new Vertex(-1 + j * xStep, -1 + i * yStep,  1);
            vB[i][j] = new Vertex(-1 + j * xStep, -1 + i * yStep, -1);
         }
      }
      // Create the triangles.
      for (int i = 0; i < yGrid + 1; i++)
      {
         for (int j = 0; j < xCount; j++)
         {
            addTriangle( vF[i  ][j  ], vF[i  ][j+1], vF[i+1][j+1] );
            addTriangle( vF[i+1][j+1], vF[i+1][j  ], vF[i  ][j  ] );
            addTriangle( vB[i  ][j  ], vB[i+1][j+1], vB[i  ][j+1] );
            addTriangle( vB[i+1][j+1], vB[i  ][j  ], vB[i+1][j  ] );
         }
      }

      // Create the right and left sides of the cube.
      // An array of vertices to be used to create the triangles.
      Vertex[][] vR = new Vertex[yGrid+2][zCount+1];
      Vertex[][] vL = new Vertex[yGrid+2][zCount+1];
      // Create all the vertices.
      for (int i = 0; i < yGrid + 2; i++)
      {
         for (int j = 0; j < zCount + 1; j++)
         {
            vR[i][j] = new Vertex( 1, -1 + i * yStep, -1 + j * zStep);
            vL[i][j] = new Vertex(-1, -1 + i * yStep, -1 + j * zStep);
         }
      }
      // Create the triangles.
      for (int i = 0; i < yGrid + 1; i++)
      {
         for (int j = 0; j < zCount; j++)
         {
            addTriangle( vR[i  ][j  ], vR[i+1][j+1], vR[i  ][j+1] );
            addTriangle( vR[i+1][j+1], vR[i  ][j  ], vR[i+1][j  ] );
            addTriangle( vL[i  ][j  ], vL[i  ][j+1], vL[i+1][j+1] );
            addTriangle( vL[i+1][j+1], vL[i+1][j  ], vL[i  ][j  ] );
         }
      }
   }
}//Cube3
