/*

*/

package renderer.pipeline;
import  renderer.scene.*;

/**
   This stage is used by the renderer to cull from a {@link Model}'s
   {@link java.util.List} of {@link Primitive}s any {@link Face} or
   {@link Triangle} that is "back facing".
<p>
   For a good explanation of orientation, forward and backward facing
   polygons, and back face culling, see Chapter 9 of<br>
      "Computer Graphics Through OpenGL: From Theory to Experiments,"
      2nd Edition, by Sumanta Guha, Chapman and Hall/CRC, 2014.
<p>
   That chapter is available online at<br>
       <a href="http://www.sumantaguha.com/files/materials/ch9.pdf" target="_top">
          http://www.sumantaguha.com/files/materials/ch9.pdf</a>
*/
public class BackFaceCulling
{
   /**
      Determine if a {@link Face} or {@link Triangle} is forward or backward
      facing. The renderer will delete from a {@link Model}'s primitive list
      any {@link Face} or {@link Triangle} that is backward facing.

      @param p                 {@link Primitive} with at least three vertices
      @param frontFacingIsCCW  determines which is front facing, CCW or CW
      @return true if the {@link Primitive} is back facing
   */
   public static boolean isBackFace(Primitive p, boolean frontFacingIsCCW)
   {
      return ((p instanceof Face) || (p instanceof Triangle))
             && (frontFacingIsCCW ^ isCCW(p));
   //https://docs.oracle.com/javase/specs/jls/se6/html/expressions.html#15.22.2
   }


   /**
      Determine if a {@link Face} or {@link Triangle} is counterclockwise.
      <p>
      This method assumes that the {@link Primitive} {@code p}
      has at least three vertices so that it can compute two
      edge vectors and then compute the z-component of their
      cross product.

      @param p  {@link Primitive} with at least three vertices
      @return true if the {@link Primitive} is counterclockwise
   */
   private static boolean isCCW(Primitive p)
   {
      // We only need to know the z-component of the cross product
      // of the two edge vectors, v1-v0 and v2-v0. This works
      // because the camera is looking down the negative z-axis.
      double zValueOfCrossProduct;

      double x0 = p.model.vertexList.get(p.indexList.get(0)).x;
      double x1 = p.model.vertexList.get(p.indexList.get(1)).x;
      double x2 = p.model.vertexList.get(p.indexList.get(2)).x;
      double y0 = p.model.vertexList.get(p.indexList.get(0)).y;
      double y1 = p.model.vertexList.get(p.indexList.get(1)).y;
      double y2 = p.model.vertexList.get(p.indexList.get(2)).y;

      double w0 = p.model.vertexList.get(p.indexList.get(0)).w;
      double w1 = p.model.vertexList.get(p.indexList.get(1)).w;
      double w2 = p.model.vertexList.get(p.indexList.get(2)).w;

      // This works in the camera's image plane, but not in
      // homogeneous clip coordinates.
      //zValueOfCrossProduct = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);

      // This works in homogeneous clip coordinates, but it has too
      // many division opperations. This is essentially the same as
      // the previous version, but with "perspective division".
      //zValueOfCrossProduct = (x1/w1 - x0/w0) * (y2/w2 - y0/w0)
      //                     - (y1/w1 - y0/w0) * (x2/w2 - x0/w0);

      // This works in homogeneous clip coordinates and it has
      // fewer division opperations.
      //zValueOfCrossProduct = (w0*x1/w1 - x0) * (w0*y2/w2 - y0)
      //                     - (w0*y1/w1 - y0) * (w0*x2/w2 - x0);

      // This works in homogeneous clip coordinates and it doesn't
      // have division opperations.
      //zValueOfCrossProduct = (w0*x1*w2 - x0*w1*w2) * (w0*w1*y2 - y0*w1*w2)
      //                     - (w0*y1*w2 - y0*w1*w2) * (w0*w1*x2 - x0*w1*w2);

      // Here is the previous version with a bit of optimization.
      //zValueOfCrossProduct = (w0*x1 - x0*w1)*w2 * (w0*y2 - y0*w2)*w1
      //                     - (w0*y1 - y0*w1)*w2 * (w0*x2 - x0*w2)*w1;

      // Some more optimization.
      zValueOfCrossProduct = ((w0*x1 - x0*w1) * (w0*y2 - y0*w2)
                            - (w0*y1 - y0*w1) * (w0*x2 - x0*w2)) * w1*w2;

      if (Pipeline.debug)
      {
         System.out.printf("zValueOfCrossProduct = % .26f\n",
                            zValueOfCrossProduct);
      }

      return 0 <= zValueOfCrossProduct;
   }
}
