/*

*/

package renderer.pipeline;
import  renderer.scene.*;

import java.util.List;
import java.util.ArrayList;

/**
   Clip a (projected) {@link Triangle} that sticks out of
   the standard homogeneous clipping region in clip coordinates..
   Interpolate {@link Vertex} color from any clipped off
   {@code Vertex} to the new {@code Vertex}.
*/
public class ClipTriangle
{
   public static boolean debug;

   /**
      If the {@link Triangle} sticks out of the homogeneous clipping region,
      then clip it so that it is contained in the homogeneous clipping region.
      <p>
      When we clip a triangle, we may end up with as many as ten
      new sub-triangles. The new, clipped, sub-triangles are
      gathered into the return {@link List}.

      @param t  {@link Triangle} to be clipped
      @return a {@link List} that is either empty or contains clipped {@code Triangle}s
   */
   public static List<Primitive> clip(Triangle t)
   {
      debug = Clip.debug; // update the debug status

      // Create an empty Primitive list to hold
      // the clipped Triangle objects.
      List<Primitive> doneList = new ArrayList<>();

      // Create an empty Triangle list to hold Triangle
      // objects that still need clipping.
      List<Triangle> toDoList = new ArrayList<>();

      toDoList.add(t);

      while ( ! toDoList.isEmpty() )
      {
         Triangle t2 = toDoList.remove(0);

         // Make local copies of several values.
         int index0 = t2.indexList.get(0);
         int index1 = t2.indexList.get(1);
         int index2 = t2.indexList.get(2);

         Vertex v0 = t2.model.vertexList.get( index0 );
         Vertex v1 = t2.model.vertexList.get( index1 );
         Vertex v2 = t2.model.vertexList.get( index2 );

         double x0 = v0.x,  y0 = v0.y, w0 = v0.w;
         double x1 = v1.x,  y1 = v1.y, w1 = v1.w;
         double x2 = v2.x,  y2 = v2.y, w2 = v2.w;

         // 1. Check for trivial accept.
         if ( ! ( Math.abs( x0 ) > Math.abs( w0 )
               || Math.abs( y0 ) > Math.abs( w0 )
               || Math.abs( x1 ) > Math.abs( w1 )
               || Math.abs( y1 ) > Math.abs( w1 )
               || Math.abs( x2 ) > Math.abs( w2 )
               || Math.abs( y2 ) > Math.abs( w2 ) ) )
         {
            // This triangle doesn't need clipping, so put it in the done list.
            if (debug) System.out.println(" Trivial accept.");
            doneList.add(t2);
         }
         // 2. Check for trivial delete.
         else if ( (w0 - x0 < 0  &&  w1 - x1 < 0  &&  w2 - x2 < 0)//right of x=w
                || (w0 + x0 < 0  &&  w1 + x1 < 0  &&  w2 + x2 < 0)//left of x=-w
                || (w0 - y0 < 0  &&  w1 - y1 < 0  &&  w2 - y2 < 0)//above y=w
                || (w0 + y0 < 0  &&  w1 + y1 < 0  &&  w2 + y2 < 0))//below y=-w
         {
            if (debug) System.out.println(" Trivial delete.");
         }
         // 3. Need to clip this triangle.
         else if (w0 - x0 < 0 || w1 - x1 < 0 || w2 - x2 < 0) //crosses x=w
         {
            // Clip to the x = w plane.
            //if (debug) System.out.println(" Clip to x = +w");
            if (w0 - x0 < 0 && w1 - x1 < 0)
            {// Create one new Triangle containing previous v2 and new v0, v1
               if (debug) System.out.println(" Clip off v0 & v1 at x = w.");
               interpolateNewTriangle(t2.model, index2, index0, index1, toDoList, 1);
            }
            else if (w0 - x0 < 0 && w2 - x2 < 0)
            {// Create one new Triangle containing previous v1 and new v0, v2.
               if (debug) System.out.println(" Clip off v0 & v2 at x = w.");
               interpolateNewTriangle(t2.model, index1, index2, index0, toDoList, 1);
            }
            else if (w1 - x1 < 0 && w2 - x2 < 0)
            {// Create one new Triangle containing previous v0 and new v1, v2.
               if (debug) System.out.println(" Clip off v1 & v2 at x = w.");
               interpolateNewTriangle(t2.model, index0, index1, index2, toDoList, 1);
            }
            else if (w0 - x0 < 0)
            {// Create two new Triangles, neither containing previous v0.
               if (debug) System.out.println(" Clip off v0 at x = w.");
               interpolateNewTriangles(t2.model, index0, index1, index2, toDoList, 1);
            }
            else if (w1 - x1 < 0)
            {// Create two new Triangles, neither containing previous v1.
               if (debug) System.out.println(" Clip off v1 at x = w.");
               interpolateNewTriangles(t2.model, index1, index2, index0, toDoList, 1);
            }
            else if (w2 - x2 < 0)
            {// Create two new Triangles, neither containing previous v2.
               if (debug) System.out.println(" Clip off v2 at x = w.");
               interpolateNewTriangles(t2.model, index2, index0, index1, toDoList, 1);
            }
            else  // We should never get here.
            {
               System.err.println("Clipping to x = +w dropped a triangle!");
               Thread.dumpStack();
               System.exit(-1);
            }
         }
         else if (w0 + x0 < 0 || w1 + x1 < 0 || w2 + x2 < 0) //crosses x=-w
         {
            // Clip to the x = -w plane.
            //if (debug) System.out.println(" Clip to x = -w");
            if (w0 + x0 < 0 && w1 + x1 < 0)
            {// Create one new Triangle containing previous v2 and new v0, v1.
               if (debug) System.out.println(" Clip off v0 & v1 at x = -w.");
               interpolateNewTriangle(t2.model, index2, index0, index1, toDoList, 2);
            }
            else if (w0 + x0 < 0 && w2 + x2 < 0)
            {// Create one new Triangle containing previous v1 and new v0, v2.
               if (debug) System.out.println(" Clip off v0 & v2 at x = -w.");
               interpolateNewTriangle(t2.model, index1, index2, index0, toDoList, 2);
            }
            else if (w1 + x1 < 0 && w2 + x2 < 0)
            {// Create one new Triangle containing previous v0 and new v1, v2.
               if (debug) System.out.println(" Clip off v1 & v2 at x = -w.");
               interpolateNewTriangle(t2.model, index0, index1, index2, toDoList, 2);
            }
            else if (w0 + x0 < 0)
            {// Create two new Triangles, neither containing previous v0.
               if (debug) System.out.println(" Clip off v0 at x = -w.");
               interpolateNewTriangles(t2.model, index0, index1, index2, toDoList, 2);
            }
            else if (w1 + x1 < 0)
            {// Create two new Triangles, neither containing previous v1.
               if (debug) System.out.println(" Clip off v1 at x = -w.");
               interpolateNewTriangles(t2.model, index1, index2, index0, toDoList, 2);
            }
            else if (w2 + x2 < 0)
            {// Create two new Triangles, neither containing previous v2.
               if (debug) System.out.println(" Clip off v2 at x = -w.");
               interpolateNewTriangles(t2.model, index2, index0, index1, toDoList, 2);
            }
            else  // We should never get here.
            {
               System.err.println("Clipping to x = -w dropped a triangle!");
               Thread.dumpStack();
               System.exit(-1);
            }
         }
         else if (w0 - y0 < 0 || w1 - y1 < 0 || w2 - y2 < 0) //crosses y=w
         {
            // Clip to the y = +w plane.
            //if (debug) System.out.println(" Clip to y = +w");
            if (w0 - y0 < 0 && w1 - y1 < 0)
            {// Create one new Triangle containing previous v2 and new v0, v1.
               if (debug) System.out.println(" Clip off v0 & v1 at y = w.");
               interpolateNewTriangle(t2.model, index2, index0, index1, toDoList, 3);
            }
            else if (w0 - y0 < 0 && w2 - y2 < 0)
            {// Create one new Triangle containing previous v1 and new v0, v2.
               if (debug) System.out.println(" Clip off v0 & v2 at y = w.");
               interpolateNewTriangle(t2.model, index1, index2, index0, toDoList, 3);
            }
            else if (w1 - y1 < 0 && w2 - y2 < 0)
            {// Create one new Triangle containing previous v0 and new v1, v2.
               if (debug) System.out.println(" Clip off v1 & v2 at y = w.");
               interpolateNewTriangle(t2.model, index0, index1, index2, toDoList, 3);
            }
            else if (w0 - y0 < 0)
            {// Create two new Triangles, neither containing previous v0.
               if (debug) System.out.println(" Clip off v0 at y = w.");
               interpolateNewTriangles(t2.model, index0, index1, index2, toDoList, 3);
            }
            else if (w1 - y1 < 0)
            {// Create two new Triangles, neither containing previous v1.
               if (debug) System.out.println(" Clip off v1 at y = w.");
               interpolateNewTriangles(t2.model, index1, index2, index0, toDoList, 3);
            }
            else if (w2 - y2 < 0)
            {// Create two new Triangles, neither containing previous v2.
               if (debug) System.out.println(" Clip off v2 at y = w.");
               interpolateNewTriangles(t2.model, index2, index0, index1, toDoList, 3);
            }
            else  // We should never get here.
            {
               System.err.println("Clipping to y = +w dropped a triangle!");
               Thread.dumpStack();
               System.exit(-1);
            }
         }
         else if (w0 + y0 < 0 || w1 + y1 < 0 || w2 + y2 < 0) //crosses y=-w
         {
            // Clip to the y = -w plane.
            //if (debug) System.out.println(" Clip to y = -w");
            if (w0 + y0 < 0 && w1 + y1 < 0)
            {// Create one new Triangle containing previous v2 and new v0, v1.
               if (debug) System.out.println(" Clip off v0 & v1 at y = -w.");
               interpolateNewTriangle(t2.model, index2, index0, index1, toDoList, 4);
            }
            else if (w0 + y0 < 0 && w2 + y2 < 0)
            {// Create one new Triangle containing previous v1 and new v0, v2.
               if (debug) System.out.println(" Clip off v0 & v2 at y = -w.");
               interpolateNewTriangle(t2.model, index1, index2, index0, toDoList, 4);
            }
            else if (w1 + y1 < 0 && w2 + y2 < 0)
            {// Create one new Triangle containing previous v0 and new v1, v2.
               if (debug) System.out.println(" Clip off v1 & v2 at y = -w.");
               interpolateNewTriangle(t2.model, index0, index1, index2, toDoList, 4);
            }
            else if (w0 + y0 < 0)
            {// Create two new Triangles, neither containing previous v0.
               if (debug) System.out.println(" Clip off v0 at y = -w.");
               interpolateNewTriangles(t2.model, index0, index1, index2, toDoList, 4);
            }
            else if (w1 + y1 < 0)
            {// Create two new Triangles, neither containing previous v1.
               if (debug) System.out.println(" Clip off v1 at y = -w.");
               interpolateNewTriangles(t2.model, index1, index2, index0, toDoList, 4);
            }
            else if (w2 + y2 < 0)
            {// Create two new Triangles, neither containing previous v2.
               if (debug) System.out.println(" Clip off v2 at y = -w.");
               interpolateNewTriangles(t2.model, index2, index0, index1, toDoList, 4);
            }
            else  // We should never get here.
            {
               System.err.println("Clipping to y = -w dropped a triangle!");
               Thread.dumpStack();
               System.exit(-1);
            }
         }
         else // We should never get here.
         {
            System.err.println("Triangle Clipping Error!");
            Thread.dumpStack();
          //System.err.println(Arrays.toString(Thread.currentThread().getStackTrace()));
            System.exit(-1);
         }
      }
      return doneList;
   }


   /**
      This method takes three indices for three vertices, one that is on
      the "right" side of a clipping plane and the other two that are on the
      "wrong" side of the clipping plane, a "todo" {@link List} into which
      a new clipped {@link Triangle} should be put (since it may need to
      be clipped again), and an integer which specifies which clipping plane
      to use when calling {@link #interpolateNewVertex}.
      <p>
      When this method returns, it will have placed in {@code triangleList}
      a clipped {@link Triangle} with the index index0_inside and the indices
      for two new interpolated vertices, both of which have been put into
      the {@link Model}'s vertex list.

      @param model           {@link Model} object that holds the vertex list
      @param index0_inside   the index of the {@link Vertex} that is inside the view rectangle
      @param index1_outside  the index of a {@link Vertex} that needs to be clipped off
      @param index2_outside  the index of a {@link Vertex} that needs to be clipped off
      @param triangleList    {@link List} to hold new {@link Triangle} object that replaces the clipped {@link Triangle}
      @param eqn_number      the identifier of the line crossed by the {@link Triangle}
   */
   private static void interpolateNewTriangle(Model model,
                                              int index0_inside,
                                              int index1_outside,
                                              int index2_outside,
                                              List<Triangle> triangleList,
                                              int eqn_number)
   {
      if (debug) System.out.printf(" Create one new triangle.\n");

      // Interpolate a new vertex between index0_inside and index1_outside.
      Vertex v1_new = interpolateNewVertex(model.vertexList.get(index0_inside),
                                           model.vertexList.get(index1_outside),
                                           eqn_number);

      // Interpolate a new vertex between index0_inside and index2_outside.
      Vertex v2_new = interpolateNewVertex(model.vertexList.get(index0_inside),
                                           model.vertexList.get(index2_outside),
                                           eqn_number);

      // Add two new vertices to the model.
      int index1_new = model.vertexList.size();
      model.addVertex(v1_new);
      model.addVertex(v2_new);
      // Create a new Triangle primitive using the new vertices.
      Triangle t_new = new Triangle(model, index0_inside, index1_new, 1+index1_new );

      // Add the new Triangle to the end of the todo list
      // since the Triangle may need to be clipped again.
      triangleList.add(t_new);

      //System.out.println( t_new );
   }


   /**
      This method takes three indices for three vertices, two that are on
      the "right" side of a clipping line and one that is on the "wrong"
      side of the clipping line, a "todo" {@link List} into which two new
      clipped {@link Triangle} objects should be put (since they may need
      to be clipped again), and an integer which specifies which clipping
      line to use when calling {@link #interpolateNewVertex}.
      <p>
      When this method returns, it will have placed in {@code triangleList}
      two new {@link Triangle} objects, one from the two inside vertices to
      a new interpolated vertex, and the other triangle from the two new
      interpolated vertices to one of the original inside vertices. Two new
      {@link Vertex} objects will have also been added to the {@link Model}'s
      vertex list.

      @param model           {@link Model} object that holds the vertex list
      @param index0_outside  the index of the {@link Vertex} that needs to be clipped off
      @param index1_inside   the index of a {@link Vertex} that is inside the view rectangle
      @param index2_inside   the index of a {@link Vertex} that is inside the view rectangle
      @param triangleList    {@link List} to hold two new {@link Triangle} objects that replace the clipped {@link Triangle}
      @param eqn_number      the identifier of the line crossed by the triangle
   */
   private static void interpolateNewTriangles(Model model,
                                               int index0_outside,
                                               int index1_inside,
                                               int index2_inside,
                                               List<Triangle> triangleList,
                                               int eqn_number)
   {
      if (debug) System.out.printf(" Create two new triangles.\n");

      // Interpolate a new vertex between v1_inside and v0_outside.
      Vertex v0_new1 = interpolateNewVertex(model.vertexList.get(index1_inside),
                                            model.vertexList.get(index0_outside),
                                            eqn_number);

      // Create a new Triangle.
      int index0_new1 = model.vertexList.size();
      model.addVertex(v0_new1);
      Triangle t1_new = new Triangle(model, index0_new1, index1_inside, index2_inside);

      // Interpolate a new vertex between v2_inside and v0_outside.
      Vertex v0_new2 = interpolateNewVertex(model.vertexList.get(index2_inside),
                                            model.vertexList.get(index0_outside),
                                            eqn_number);

      // Create another new Triangle.
      int index0_new2 = model.vertexList.size();
      model.addVertex(v0_new2);
      Triangle t2_new = new Triangle(model, index0_new1, index0_new2, index2_inside);

      // Add the new Triangles to the end of the todo list
      // since the Triangles may need to be clipped again.
      triangleList.add(t1_new);
      triangleList.add(t2_new);

      //System.out.println( t1_new );
      //System.out.println( t2_new );
   }

   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping plane and the other that is on the "wrong" side of the
      clipping plane, and an integer which specifies which clipping plane
      to use, where
      <pre>{@code
         eqn_number == 1 means clipping plane x =  w
         eqn_number == 2 means clipping plane x = -w
         eqn_number == 3 means clipping plane y =  w
         eqn_number == 4 means clipping plane y = -w
      }</pre>
      This method returns the vertex that is the intersection point
      between the given line segment and the given clipping plane.
      <p>
      This method solves for the value of {@code t} for which the
      parametric equation
      <pre>{@code
                  p(t) = (1-t) * v_outside + t * v_inside
      }</pre>
      intersects the given clipping plane. (Notice that the equation
      is parameterized so that we move from the outside vertex towards
      the inside vertex as {@code t} increases from 0 to 1.) The solved
      for value of {@code t} is then plugged into the parametric formula
      to get the coordinates of the intersection point.

      @param v_inside    the Vertex that is inside the clipping region
      @param v_outside   the Vertex that is outside the clipping region
      @param eqn_number  the identifier of the plane crossed by the triangle edge
      @return a Vertex object that replaces the clipped off vertex
   */
   private static Vertex interpolateNewVertex(Vertex v_inside,
                                              Vertex v_outside,
                                              int eqn_number)
   {
      //if (debug) System.out.println(" Create new vertex.");

      // Make local copies of several values.
      double vix =  v_inside.x; // "i" for "inside"
      double viy =  v_inside.y;
      double viz =  v_inside.z;
      double viw =  v_inside.w;
      double vox = v_outside.x; // "o" for "outside"
      double voy = v_outside.y;
      double voz = v_outside.z;
      double vow = v_outside.w;
      // Interpolate between v_outside and v_inside.
      double t = 0.0;
      if (1 == eqn_number)             // clip to plane x = w
         t = -(vow - vox)/( (viw - vix) - (vow - vox) );
      else if (2 == eqn_number)        // clip to plane x = -w
         t = -(vow + vox)/( (viw + vix) - (vow + vox) );
      else if (3 == eqn_number)       // clip to plane y = w
         t = -(vow - voy)/( (viw - viy) - (vow - voy) );
      else if (4 == eqn_number)       // clip to plane y = -w
         t = -(vow + voy)/( (viw + viy) - (vow + voy) );

      //if (t < 0.9999999)
      //t = t + 0.0000001;  /*** keep the new vertex off the edge!! ***/

      // Use the value of t to interpolate the coordinates of the new vertex.
    //double x = (1-t) * vox + t * vix;
    //double y = (1-t) * voy + t * viy;
    //double z = (1-t) * voz + t * viz;
    //double w = (1-t) * vow + t * viw;
      double x = vox + t * (vix - vox);
      double y = voy + t * (viy - voy);
      double z = voz + t * (viz - voz);
      double w = vow + t * (viw - vow);

      Vertex v_new = new Vertex(x, y, z, w);

      // Use the value of t to interpolate the color of the new vertex.
    //double r = (1-t) * v_outside.r + t * v_inside.r;
    //double g = (1-t) * v_outside.g + t * v_inside.g;
    //double b = (1-t) * v_outside.b + t * v_inside.b;
      double r = v_outside.r + t * (v_inside.r - v_outside.r);
      double g = v_outside.g + t * (v_inside.g - v_outside.g);
      double b = v_outside.b + t * (v_inside.b - v_outside.b);

      if (debug)
      {
         System.out.printf("  t = % .25f\n", t);
         System.out.printf("  <x_o,y_o,z_o,w_o> = <% .24f % .24f\n", vox,voy);
         System.out.printf("                       % .24f % .24f>\n",voz,vow);
         System.out.printf("  <x_i,y_i,z_i,w_i> = <% .24f % .24f\n", vix,viy);
         System.out.printf("                       % .24f % .24f>\n",viz,viw);
         System.out.printf("  <x,  y,  z,  w>   = <% .24f % .24f\n",   x,  y);
         System.out.printf("                       % .24f % .24f>\n",  z,  w);
         System.out.printf("  <r_o,g_o,b_o> = <% .15f  % .15f  % .15f>\n",
                              v_outside.r, v_outside.g, v_outside.b);
         System.out.printf("  <r_i,g_i,b_i> = <% .15f  % .15f  % .15f>\n",
                              v_inside.r, v_inside.g, v_inside.b);
         System.out.printf("  <r,  g,  b>   = <% .15f  % .15f  % .15f>\n",
                               r,  g,  b);
      }

      v_new.setColor((float)r, (float)g, (float)b);

      return v_new;
   }
}
