/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
    This renderer takes as its input a {@link Scene} data structure
    and a {@link FrameBuffer} data structure. This renderer mutates
    the {@link FrameBuffer}'s current viewport so that it is filled
    in with the rendered image of the scene represented by the
    {@link Scene} object.
<p>
    This implements our thirteenth rendering pipeline. This pipeline
    splits the projection stage into two stages, a linear projection
    transformation (that uses matrix multiplication) and a non-linear
    perspective division. The clipping stage is moved between these
    two parts of the projection operation. This puts clipping before
    the non-linear part of projection, and this solves a problem we
    had with linear interpolation (lerp) in the clipping stage.
    There are ten pipeline stages
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position(scene, position2, Matrix.identity(), fb);
         }
      }
   }


   /**
      Recursively renderer a {@link Position} without any debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-world transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 1. Apply the current model-to-world coordinate transformation.
         Model2World.model2world(position.model.vertexList, ctm);

         // 2. Apply the camera's world-to-view coordinate transformation.
         World2View.world2view(position.model.vertexList, scene.camera.viewMatrix);

         // 3. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(position.model.vertexList, scene.camera.normalizeMatrix);

         // 4. Apply the projection transformation to homogeneous clip coordinates.
         Projection.project(position.model.vertexList, scene.camera.projMatrix);

         // 5. Do back face culling (in homogeneous coordinates).
         if (position.model.doBackFaceCulling)
         {
            List<Primitive> primitiveList = new ArrayList<>();
            for (Primitive p : position.model.primitiveList)
            {
               if (! BackFaceCulling.isBackFace(p, position.model.frontFacingIsCCW))
               {
                  // Keep the primitives with the desired orientation.
                  primitiveList.add(p);
               }
            }
            position.model.primitiveList = primitiveList;
         }

         // 6. Do the primitive assembly operation.
         PrimitiveAssembly.assemble(position.model);

         // 7. Clip each primitive (in homogeneous clip space).
         List<Primitive> primitiveList = new ArrayList<>();
         for (Primitive p : position.model.primitiveList)
         {
            primitiveList.addAll( Clip.clip(p) );
         }
         position.model.primitiveList = primitiveList;

         // 8. Apply the perspective division operation to NDC.
         PerspectiveDivision.perspectiveDivide(position.model.vertexList);

         // 9. Apply viewport transformation to each visible primitive.
         // (Alternatively, we could apply viewport transformation to every vertex?)
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (Primitive p : position.model.primitiveList)
         {
            Viewport.viewport(p, transformed, fb);
         }

         // 10. Rasterize each visible primitive into pixels.
         for (Primitive p : position.model.primitiveList)
         {
            Rasterize.rasterize(p, fb);
         }

         // Recursively render every nested Position of this Model.
         for (Position p : position.model.nestedPositions)
         {
            if ( p.visible )
            {
               // Do a (pre-order) depth-first-traversal from this nested Position.
               render_position(scene, p, ctm, fb);
            }
         }
      }

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position(scene, p, ctm, fb);
         }
      }
   }



   /**
      Renderer the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         System.out.println("=== Render Position ================================");
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position_debug(scene, position2, Matrix.identity(), fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("=== End Position ================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a {@link Position} with quite a bit of debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-world transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position_debug(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      System.out.println("=== Render Model ===================================");

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // log interesting information to standard output
         logVertexList("0. Model", position.model.vertexList);

         // 1. Apply the current model-to-world coordinate transformation.
         Model2World.model2world(position.model.vertexList, ctm);

         // log interesting information to standard output
         logVertexList("1. World", position.model.vertexList);

         // 2. Apply the camera's world-to-view coordinate transformation.
         World2View.world2view(position.model.vertexList, scene.camera.viewMatrix);

         // log interesting information to standard output
         logVertexList("2. View", position.model.vertexList);

         // 3. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(position.model.vertexList, scene.camera.normalizeMatrix);

         // log interesting information to standard output
         logVertexList("3. Camera", position.model.vertexList);

         // 4. Apply the projection transformation to homogeneous clip coordinates.
         Projection.project(position.model.vertexList, scene.camera.projMatrix);

         // log interesting information to standard output
         logVertexList("4. Projected", position.model.vertexList);

         // 5. Do back face culling (in homogeneous coordinates).
         if (position.model.doBackFaceCulling)
         {
            List<Primitive> primitiveList = new ArrayList<>();
            for (Primitive p : position.model.primitiveList)
            {
               if (! BackFaceCulling.isBackFace(p, position.model.frontFacingIsCCW))
               {
                  // Keep the primitives with the desired orientation.
                  primitiveList.add(p);
               }
            }
            position.model.primitiveList = primitiveList;
         }

         // log interesting information to standard output
         logPrimitiveList("5. Culling", position.model.primitiveList);

         // 6. Do the primitive assembly operation.
         PrimitiveAssembly.assemble(position.model);

         // log interesting information to standard output
         logPrimitiveList("6. Assembly", position.model.primitiveList);

         // 7. Clip each primitive (in homogeneous clip space).
         List<Primitive> primitiveList = new ArrayList<>();
         for (Primitive p : position.model.primitiveList)
         {
            // log interesting information to standard output
            logPrimitive("7. Clipping", p);

            List<Primitive> clipped = Clip.clip(p);
            if ( 0 != clipped.size() )
            {
               for (Primitive p2 : clipped)
               {
                  // log interesting information to standard output
                  logPrimitive("7. Clipped (accept)", p2);

                  // Keep the primitives that are visible.
                  primitiveList.add(p2);
               }
            }
            else
            {
               // log interesting information to standard output
               logPrimitive("7. Clipped (reject)", p);
            }
         }
         position.model.primitiveList = primitiveList;

         // log interesting information to standard output
         logVertexList("7. Clipped", position.model.vertexList);

         // 8. Apply the perspective division operation to NDC.
         PerspectiveDivision.perspectiveDivide(position.model.vertexList);

         // log interesting information to standard output
         logVertexList("8. NDC", position.model.vertexList);

         // 9. Apply viewport transformation to each visible primitive.
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (Primitive p : position.model.primitiveList)
         {
            Viewport.viewport(p, transformed, fb);

            // log interesting information to standard output
            logPrimitive("9. Viewport", p);
         }

         // 10. Rasterize each visible primitive into pixels.
         for (Primitive p : position.model.primitiveList)
         {
            // log interesting information to standard output
            logPrimitive("10. Rasterize", p);

            Rasterize.rasterize(p, fb);
         }
         System.out.println();

         // Recursively render every nested Position of this Model.
         for (Position position2 : position.model.nestedPositions)
         {
            System.out.println("====== Render Model's Nested Position ================");
            if ( position2.visible )
            {
               // Do a pre-order depth-first-traversal from this nested Position.
               render_position(scene, position2, ctm, fb);
            }
            else
            {
               System.out.println("   Hidden position.");
            }
            System.out.println("====== End Model's Nested Position ===================");
         }
      }
      else
      {
         System.out.println("   Hidden model.");
      }

      System.out.println("=== End Model ===================================");

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         System.out.println("====== Render Nested Position ========================");
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position_debug(scene, p, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("====== End Nested Position ===========================");
      }
   }


   /**
      This method prints a {@link String} representation of the given
      vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Primitive} list.

      @param stage          name for the pipeline stage
      @param primitiveList  primitive list whose string representation is to be printed
   */
   private static void logPrimitiveList(String stage, List<Primitive> primitiveList)
   {
      for (Primitive p : primitiveList)
      {
         System.out.printf( stage + ": " + p.toString() );
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link Primitive}.

      @param stage  name for the pipeline stage
      @param p      {@link Primitive} whose string representation is to be printed
   */
   private static void logPrimitive(String stage, Primitive p)
   {
      System.out.print( stage + ": " + p.toString() );
      for (Integer index : p.indexList)
      {
         Vertex v = p.model.vertexList.get(index);
         System.out.printf("    index = %3d, " + v.toString(), index);
         System.out.printf("                 " + v.toStringColor());
      }
   }
}
