/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;

/**
   A {@code Face} object has a list of integers that represent
   the endpoints of a closed loop of line  segments. The
   integers are indices into the {@link Vertex} list of a
   {@link Model} object.
<p>
   A {@code Face} is closed loop of line segments that is assumed
   to be "flat" and "convex".
<p>
   Flat means that the vertices of the {@code Face} all lie in
   the same plane.
<p>
   Convex means that a line segment joining any two vertices of
   the {@code Face} must be completely contained inside of the
   {@code Face}.
<p>
   The combination of flat and convex means that we can give a
   {@code Face} an orientation and we can determine if the
   {@code Face} if "front facing" or "back facing".
*/
public class Face extends Primitive
{
   /**
      Construct a default (empty) {@code Face}.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Face}
   */
   public Face(Model model)
   {
      super(model);
   }


   /**
      Construct a {@code Face} from an array of integer indices.

      @param model   {@link Model} containing the {@link Vertex} list for the new {@code Face}
      @param iArray  array of vertices to put into the new {@code Face}
   */
   public Face(Model model, Integer... iArray)
   {
      super(model, iArray);
   }


   /**
      Construct a {@code Face} from a {@link List} of integer indices.

      @param model  {@link Model} containing the {@link Vertex} list for the new {@code Face}
      @param iList  {@link List} of vertices to put into the new {@code Face}
   */
   public Face(Model model, List<Integer> iList)
   {
      super(model, iList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this {@code Face} primitive.

      @param newModel  the {@link Model} that the new {@code Face} will be part of
      @return a {@code Face} object that is a deep copy of this {@code Face}
   */
   public Face makeCopy(Model newModel)
   {
      List<Integer> newIndexList = new ArrayList<>(this.indexList.size());
      for (Integer index : this.indexList)
      {
         newIndexList.add(index);
      }
      return new Face(newModel, newIndexList);
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Face} object
   */
   @Override
   public String toString()
   {
      String result = "Face: (";
      for (Integer index : this.indexList)
      {
         result += index + ", ";
      }
      return result + ")\n";
   }
}
