/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   A {@code Model} data structure represents a distinct geometric object
   in a {@link Scene}. A {@code Model} data structure is a {@link List}
   of {@link Vertex} object and another list of {@link Primitive} objects.
   {@link Primitive} objects represents "geometric primitives" that the
   model's geometry is build out of. There are three kinds of geometric
   primitives, line segments, faces, and triangles. Each {@link LineSegment}
   primitive contains two integers that are indices of two {@link Vertex}
   objects from the {@code Models}'s vertex list. The two {@link Vertex}
   objects contain the coordinates, in the model's local coordinate system,
   of the line segment's two endpoints. A {@link Face} primitive should be
   a closed, flat, convex polygon. Each {@link Face} primitive contains
   a list of integer indices of {@link Vertex} objects from the {@code Models}'s
   vertex list. The vertices are the corners of the face's polygon. A Face
   is rendered as a sequence of line segments. A {@link Triangle} primitive
   is a three sided face, but it is rendered as a solid, filled in, triangle.
<p>
   A {@code Model} is also a recursive data structure. Every {@code Model}
   object contains a {@link List} of {@link Position} objects. The list of
   {@link Position}s lets us define what are called hierarchical, or nested,
   models. These are models that are made up of distinct parts and the
   distinct parts need to move both individually and also move together as
   a group. Each {@link Position} object nested inside of a {@code Model}
   contains its own {@link Matrix} which determines a local coordinate
   system for the vertices in that {@link Position}'s {@link Model} and
   for all the other {@link Model} objects that might be below that
<p>
   Each nested {@link Position}'s {@link Matrix} will place the nested
   {@link Position}'s {@link Model} within its parent's local coordinate
   system. Another way to say this is that the {@link Matrix} of a nested
   {@link Position} determines its {@link Model}'s location and orientation
   within the local coordinate system of the nested {@link Position}'s
   parent {@link Position}. In addition, the nested {@link Position}'s
   {@link Matrix} determines a new local coordinate system within which all
   the nested {@link Position}s below a {@link Position} will place their {@link Model}s
<p>

   The "top level" Position's matrix will place the whole model within the
   camera's coordinate system).




*/
public class Model
{
   public boolean doBackFaceCulling = true;
   public boolean frontFacingIsCCW  = true;

   public List<Vertex> vertexList = new ArrayList<>();
   public List<Primitive> primitiveList = new ArrayList<>();

   public List<Position> nestedPositions = new ArrayList<>();

   public boolean visible;


   /**
      Construct an empty {@code Model} object.
   */
   public Model()
   {
      visible = true;
      doBackFaceCulling = true;
      frontFacingIsCCW = true;
   }


   /**
      A "copy constructor". This constructor should make a deep copy
      of the given {@code Model}'s {@link Vertex} list, {@link LineSegment}
      list, and {@link List} of nested {@link Position}s.

      @param model  {@code Model} to make a copy of
   */
   public Model(Model model) // a "copy constructor"
   {
      super();

      this.visible = model.visible;
      this.doBackFaceCulling = model.doBackFaceCulling;
      this.frontFacingIsCCW = model.frontFacingIsCCW;
      for (Vertex v : model.vertexList)
      {
         this.vertexList.add(new Vertex(v)); // deep copy of each Vertex
      }
      for (Primitive p : model.primitiveList)
      {
         this.primitiveList.add(p.makeCopy(this));
      }
      //this.nestedPositions = new ArrayList<>();
      for (Position p : model.nestedPositions)
      {
         this.nestedPositions.add(new Position(p));
      }
   }


   /**
      Add a nested {@link Position} (or Positions) to this {@code Model}'s
      {@link List} of nested {@link Position}s.

      @param pArray  array of nested {@link Position}s to add to this {@code Model}
   */
   public void addNestedPosition(Position... pArray)
   {
      for (Position p : pArray)
      {
         this.nestedPositions.add(p);
      }
   }


   /**
      Get the index to a nested {@link Position} from this {@code Model}'s
      {@link List} of nested {@link Position}s.

      @param p  {@link Position} to search for in this {@code Model}'s {@link List} of nested {@link Position}s
      @return the index of the first occurrence of the specified {@link Position},
              or -1 if the {@link Position} list does not contain the {@link Position}
   */
   public int indexOfNestedPosition(Position p)
   {
      return this.nestedPositions.indexOf(p);
   }


   /**
      Get a reference to the nested {@link Position} at the given index in this {@code Model}'s
      {@link List} of nested {@link Position}s.

      @param index  index of the nested {@link Position} to return
      @return nested {@link Position} at the specified index in the {@link List} of nested {@link Position}s
      @throws IndexOutOfBoundsException if the index is out of range
              {@code (index < 0 || index >= size())}
   */
   public Position getNestedPosition(int index)
   {
      return nestedPositions.get(index);
   }


   /**
      Replace the nested {@link Position} at the given index in this {@code Model}'s
      {@link List} of nested {@link Position}s.

      @param index  index of the nested {@link Position} to replace
      @param p      {@link Position} to be stored at the specified index
      @return {@link Position} previously at the specified index
      @throws IndexOutOfBoundsException if the index is out of range
              {@code (index < 0 || index >= size())}
   */
   public Position setNestedPosition(int index, Position p)
   {
      return this.nestedPositions.set(index, p);
   }


   /**
      Add a {@link Vertex} (or vertices) to this {@code Model}'s
      {@link List} of vertices.

      @param vArray  array of {@link Vertex} objects to add to this {@code Model}
   */
   public void addVertex(Vertex... vArray)
   {
      for (Vertex v : vArray)
      {
         this.vertexList.add(new Vertex(v)); // NOTE: deep copy!
      }
   }


   /**
      Add a {@link Primitive} (or primitives) to this {@code Model}'s
      {@link List} of primitives.

      @param pArray  array of {@link Primitive} objects to add to this {@code Model}
   */
   public void addPrimitive(Primitive... pArray)
   {
      for (Primitive p : pArray)
      {
         this.primitiveList.add(p);
      }
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param i0  1st integer index of {@link LineSegment} to add to this {@code Model}
      @param i1  2nd integer index of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(int i0, int i1)
   {
      this.primitiveList.add(new LineSegment(this, i0, i1));
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      This method adds (copies of) the two give {@link Vertex} objects to
      this {@code Model}'s vertex list and then costructs an appropriate
      {@link LineSegment} object to add to this {@code Model}'s primitive
      list.
      <p>
      This method provides a convenient way to build up a {@code Model} but
      it may not build an efficient {@code Model}. Using this mehtod may lead
      to a lot of redundant {@link Vertex} objects in this {@code Model}'s
      vertex list.

      @param v0  1st {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param v1  2nd {@link Vertex} of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(Vertex v0, Vertex v1)
   {
      int index = this.vertexList.size();
      this.vertexList.add(new Vertex(v0)); // NOTE: deep copy!
      this.vertexList.add(new Vertex(v1)); // NOTE: deep copy!
      this.primitiveList.add(new LineSegment(this, index, index+1));
   }


   /**
      Add a {@link Face} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param iArray  array of integer indices for the {@link Face}
                     to add to this {@code Model}
   */
   public void addFace(Integer... iArray)
   {
      this.primitiveList.add(new Face(this, iArray));
   }


   /**
      Add a {@link Face} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      This method adds (copies of) the give {@link Vertex} objects to
      this {@code Model}'s vertex list and then costructs an appropriate
      {@link Face} object to add to this {@code Model}'s primitive list.
      <p>
      This method provides a convenient way to build up a {@code Model} but
      it may not build an efficient {@code Model}. Using this mehtod may lead
      to a lot of redundant {@link Vertex} objects in this {@code Model}'s
      vertex list.

      @param vArray  array of {@link Vertex} objects for the {@link Face}
                     to add to this {@code Model}
   */
   public void addFace(Vertex... vArray)
   {
      List<Integer> indexList = new ArrayList<>(vArray.length);
      int index = this.vertexList.size();
      for (Vertex v : vArray)
      {
         this.vertexList.add(new Vertex(v)); // NOTE: deep copy!
         indexList.add(index);
         index++;
      }
      this.primitiveList.add(new Face(this, indexList));
   }


   /**
      Add a {@link Triangle} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param iArray  array of integer indices for the {@link Triangle}
                     to add to this {@code Model}
   */
   public void addTriangle(Integer... iArray)
   {
      this.primitiveList.add(new Triangle(this, iArray));
   }


   /**
      Add a {@link Triangle} to this {@code Model}'s {@link List} of
      {@link Primitive}s.
      <p>
      This method adds (copies of) the give {@link Vertex} objects to
      this {@code Model}'s vertex list and then costructs an appropriate
      {@link Triangle} object to add to this {@code Model}'s primitive list.
      <p>
      This method provides a convenient way to build up a {@code Model} but
      it may not build an efficient {@code Model}. Using this mehtod may lead
      to a lot of redundant {@link Vertex} objects in this {@code Model}'s
      vertex list.

      @param vArray  array of {@link Vertex} objects for the {@link Triangle}
                     to add to this {@code Model}
   */
   public void addTriangle(Vertex... vArray)
   {
      List<Integer> indexList = new ArrayList<>(vArray.length);
      int index = this.vertexList.size();
      for (Vertex v : vArray)
      {
         this.vertexList.add(new Vertex(v)); // NOTE: deep copy!
         indexList.add(index);
         index++;
      }
      this.primitiveList.add(new Triangle(this, indexList));
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to the same {@link Color}.

      @param c  {@link Color} for all of this model's {@link Vertex} objects
   */
   public void setColor(Color c)
   {
      for (Vertex v : this.vertexList)
      {
         v.setColor(c);
      }
      // Recursively set the color for the nested models.
      for (Position p : this.nestedPositions)
      {
         p.model.setColor(c);
      }
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to the same random {@link Color}.
   */
   public void setColorRandom()
   {
      Vertex v0 = new Vertex();
      v0.setColorRandom();
      Color c = v0.getColor();
      for (Vertex v : this.vertexList)
      {
         v.setColor(c);
      }
      // Recursively set the color for the nested models.
      for (Position p : this.nestedPositions)
      {
         p.model.setColor(c);
      }
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to a different random {@link Color}.
   */
   public void setRandomColors()
   {
      for (Vertex v : this.vertexList)
      {
         v.setColorRandom();
      }
      // Recursively set the color for the nested models.
      for (Position p : this.nestedPositions)
      {
         p.model.setRandomColors();
      }
   }


   /**
      Set each {@link Primitive} in this {@code Model}
      to a different random {@link Color}.
      <p>
      NOTE: This works best when the {@link Primitive}
      objects in this {@link Model} do not share any
      {@link Vertex} objects.
   */
   public void setRandomPrimitveColors()
   {
      for (Primitive p : this.primitiveList)
      {
         p.setColorRandom();
      }
      // Recursively set the color for the nested models.
      for (Position p : this.nestedPositions)
      {
         p.model.setRandomPrimitveColors();
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Model} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Model has " + vertexList.size() + " vertices.\n";
      result += "Model has " + primitiveList.size() + " primitves.\n";
      int i = 0;
      for (Vertex v : this.vertexList)
      {
         result += i + ": " + v.toString();
         i++;
      }
      //result = "Printing out this Model's " + primitiveList.size() + " Primitives:\n";
      for (Primitive p : this.primitiveList)
      {
         result += p.toString();
      }
      result += "This Model has " + nestedPositions.size() + " nested Positions\n";
      for (Position p : this.nestedPositions)
      {
         result += p.toString();
      }
      //result += "Done printing out Model\n";
      return result;
   }
}
