/*

*/

package renderer.scene;

/**
   This is a perspective projection {@link Matrix}.
   <pre>{@code
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  0  1 ]
     [ 0  0 -1  0 ]
   }</pre>
   Notice that this matrix maps
   <pre>{@code
      (x, y, z, 1) -> (x, y, 1, -z)
   }</pre>
   so it "projects" the three-dimensional camera space {@code w = 1}
   onto the three-dimensional homogeneous space {@code z = 1} in clip
   coordinates..
<p>
   After the perspective division stage, points {@code (x, y, z, 1)}
   from camera space become
   <pre>{@code
      (x, y, z, 1) -> (x, y, 1, -z) -> (-x/z, -y/z, -1/z, 1)
   }</pre>
   so we will have implemented our original projection formulas
   for the {@code x} and {@code y} coordinates. And we have some
   "distorted" depth information in the z-coordinate. This depth
   information is called "pseudo-depth" or "hyperbolic depth".
<p>
   We have just shown that the above projection matrix does two things
   for us. First, when combined with perspective division it implements
   the standard projection formulas for the {@code x} and {@code y}
   coordinates. Second, it preserves some depth information. But this
   projection matrix also does a third thing for us. It breaks the
   projection transformation into two steps, a linear step (the matrix
   transformation) and a non-linear step (the perspective division). We
   will show later that if we move our clipping stage to be in the middle
   of these two projection steps, then we will have solved a problem with
   linear interpolation in the clipping stage.
<p>
   Next, we want to show that this matrix, followed by perspective division,
   transforms the standard view pyramid in camera coordinates into the
   standard infinite parallelepiped parallel to the z-axis used by the
   orthographic projection (in normalized device coordinates).
<p>
   Consider a projective ray from the origin in camera space through
   a point {@code (x, y, -1, 1)} in the camera's view plane.
   <pre>{@code
       (x*t, y*t, -t, 1)  with 0 < t < infinity
   }</pre>
   Here is what this ray is transformed into by the matrix and perspective
   division.
   <pre>{@code
      (x*t, y*t, -t, 1) -> (x*t, y*t, 1,   t)  (clip coordinates)
                        -> (x,   y,   1/t, 1)  (normalized device coordinates)
   }</pre>
   The transformed ray is (part of) a line in NDC parallel to the z-axis
   and that intersects the plane {@code z = 1} at {@code (x, y, 1, 1)}
   when {@code t=1}.
<p>
   So the combination of this matrix transformation and perspective
   division sends projective rays from camera space to parallel horizontal
   lines in normalized device coordinates. Notice how this transformation
   distorts the standard projective view volume. At the camera end of the
   perspective view pyramid this transformation "expands" all the projective
   rays at the apex of the view pyramid into the standard view rectangle
   {@code -1 <= x <= 1} and {@code -1 <= y <= 1}. At the far end of the
   standard perspective view pyramid it "compresses" the projective rays so
   that they all fit into, again, the standard view rectangle. So the
   standard infinite perspective view pyramid in camera coordinates gets
   sent by this transformation to the standard infinite view parallelepiped
   in NDC. Also, the "squeezing" of the far end of the view pyramid into
   the standard view rectangle is how this transformation implements the
   foreshortening effect of perspective viewing where distant objects appear
   smaller when projected into the view plane.
<p>
   Since a solid, three-dimensional, view pyramid is sent to a solid,
   three-dimensional, view parallelepiped, the projection transformation
   based on this matrix preserves depth information, which was one of our
   goals for this renderer.
<p>
   Finally, notice that this view parallelepiped is the same infinite view
   parallelepiped that is used by the orthographic projection. This is
   important. The orthographic projection and the perspective projection
   each transform their very different view volumes (in camera coordinates)
   into the exact same view volume in NDC. All the stages of the pipeline
   after perspective division will work the same for either orthographically
   or perspectively projected scenes.
*/
public class PerspectiveProjectionMatrix
{
   /**
      This is a static factory method.
      <p>
      Construct this perspective projection {@link Matrix}.
      <pre>{@code
        [ 1  0  0  0 ]
        [ 0  1  0  0 ]
        [ 0  0  0  1 ]
        [ 0  0 -1  0 ]
      }</pre>
   */
   public static Matrix build()
   {
      return Matrix.build(
               new Vector(1.0, 0.0, 0.0,  0.0),
               new Vector(0.0, 1.0, 0.0,  0.0),
               new Vector(0.0, 0.0, 0.0, -1.0),
               new Vector(0.0, 0.0, 1.0,  0.0) );
   }
}

/*
   Exercise: Replace the projection matrix with this one.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  0  0 ]
     [ 0  0 -1  0 ]
  Explain the results.
  (Remember that the vectors in the Matrix constructor
  are COLUMN vectors, even though they look in the code
  like row vectors.
*/

/*
   Exercise: Replace the projection matrix with this one.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  0 -1 ]
     [ 0  0 -1  0 ]
  Explain the results.
*/
